import asyncio
import aiofiles
import termios
import sys
import os

LOCAL_FLAG = 3  # [iflag, oflag, cflag, lflag, ispeed, ospeed, cc]


class Console:
    def __init__(self, cmd_cb):
        self.s = ""
        self.history = []
        self.hp = len(self.history)  # history pointer
        self.commands = []
        self.cursor_pos = 0

        self.cmd_cb = cmd_cb
        self.temp_s = None
        self.prompt = "shella$ "
        self.debug = False
        self.update_line()
        sys.stdout.flush()

    async def read_input_stream(self):
        fd = sys.stdin.fileno()
        previous_flags = termios.tcgetattr(fd)
        all_flags = termios.tcgetattr(fd)

        all_flags[LOCAL_FLAG] &= ~termios.ECHO
        all_flags[LOCAL_FLAG] &= ~termios.ICANON

        termios.tcsetattr(fd, termios.TCSANOW, all_flags)
        try:
            async with aiofiles.open(os.ttyname(fd), mode='r') as f:
                while True:
                    c = await f.read(1)
                    if c == '\x1b':
                        char = c

                        """
                        We just want to receive the whole escaped character, so lets check when it ends, 
                        i.e. when the command character such as 'A' for up arrow
                        """
                        for i in range(20):  # Dont get stuck if we screw up
                            c = await f.read(1)
                            char += c
                            if c.isnumeric() or c in [';', '[', '?', '=']:
                                continue
                            else:
                                break
                        c = char
                    await self.on_input_character(c)
        except (KeyboardInterrupt, asyncio.exceptions.CancelledError):
            pass
        except:
            import traceback
            traceback.print_exc()

        finally:
            termios.tcsetattr(fd, termios.TCSANOW, previous_flags)

    def stdout(self, c, flush=False):
        sys.stdout.write(c)
        if flush:
            sys.stdout.flush()

    def update_line(self):
        self.stdout('\x1b[2K\033[1G')
        self.print_prompt()
        self.stdout(self.s)
        for i in range(len(self.s)-self.cursor_pos):
            self.stdout('\x1b[D')

    def print_prompt(self):
        # self.stdout(f"\033[92m PROMPT [{self.cursor_pos:2}\t{self.hp}/{len(self.history)}]$ \033[0m")
        if self.debug:
            self.stdout(
                f"\033[92m{self.prompt} [{self.cursor_pos:2}\t{self.hp}/{len(self.history)}]$ \033[0m")
        else:
            self.stdout(f"\033[92m{self.prompt}\033[0m")

    def print_hist(self, dir_):

        if len(self.history) == 0:
            return
        # Note addidtion of extra empty slot at end of history
        max_len = len(self.history)

        self.hp += dir_
        if self.hp <= 0:
            self.hp = 0

        if self.hp >= max_len:
            self.hp = max_len

        self.s = (self.history+[self.temp_s])[self.hp]
        self.cursor_pos = len(self.s)

        self.update_line()

    async def on_input_character(self, x):
        if x == '\x7f':
            if (len(self.s) != 0) and (self.cursor_pos > 0):
                self.s = self.s[0:self.cursor_pos-1]+self.s[self.cursor_pos:]
                self.cursor_pos = len(self.s[0:self.cursor_pos-1])
                self.update_line()
        elif x == '\x1b[3~':  # delete
            if (len(self.s) != 0) and (self.cursor_pos >= 0):
                self.s = self.s[0:self.cursor_pos]+self.s[self.cursor_pos+1:]
                self.update_line()
        elif x in ['\x03', '\x1a']:
            return
        elif x == '\x1b[A':  # up
            self.print_hist(-1)
        elif x == '\x1b[B':  # down
            self.print_hist(1)
        elif x == '\x1b[D':  # left
            if self.cursor_pos > 0:
                self.cursor_pos -= 1
                self.stdout(x)
        elif x == '\x1b[C':  # right
            if self.cursor_pos < len(self.s):
                self.cursor_pos += 1
                self.stdout(x)
            pass

        elif x == '\n':
            if len(self.s) > 0:
                print()
                await self.cmd_cb(self.s.split())
            else:
                print("")
            valid_for_history = True
            if len(self.history) > 0:
                if self.s == self.history[-1]:
                    valid_for_history = False

            if len(self.s) > 0 and valid_for_history:
                self.history.append(self.s)
            self.temp_s = ""

            self.hp = len(self.history)
            self.s = ""
            self.cursor_pos = 0
            self.print_prompt()
        elif x == '\t':
            L = len(self.s)
            hits = 0
            options = []
            for c in self.commands:
                if self.s == c[0:L]:
                    hits += 1
                    autocomplete = c[L:]
                    options.append(c)

            if hits == 1:

                self.stdout(autocomplete)
                self.s += autocomplete
                self.cursor_pos = len(self.s)
            elif hits != 0:
                print('\033[2K\033[1G'+",".join(options))
                self.print_prompt()
                self.stdout(self.s)
                self.cursor_pos = len(self.s)
        else:

            if x.isprintable():
                self.s = self.s[0:self.cursor_pos]+x+self.s[self.cursor_pos:]
                self.temp_s = self.s
                self.cursor_pos += 1
                self.update_line()

        sys.stdout.flush()

    def set_prompt(self, prompt):
        self.prompt = prompt
        self.update_line()



import asyncio
from inspect import signature
from shella.term import Console

_cmds = []


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


async def run_command(argv):
    found = False
    for cmd in _cmds:
        if argv[0] in cmd.get("commands"):

            arg_template = cmd.get("arg_template")
            if arg_template:
                if not check_string_format(" ".join(argv[1:]), arg_template):
                    error(
                        f"Incorrect arguments, please provide on the format {arg_template}")
                    error(f"Command usage: {argv[0]} {cmd.get('usage')}")

                    return

            callback = cmd.get("callback")
            try:
                await callback(argv)
            except Exception:
                import traceback
                print("Command error")
                traceback.print_exc()
            found = True
    if not found:
        help()

c = Console(run_command)


def error(s):
    print(f"{bcolors.FAIL}ERROR: {s}{bcolors.ENDC}")


"""
Function decorator
"""


def shell_cmd(*args, **kwargs):
    def inner(func):
        o = {
        }

        o["commands"] = kwargs["commands"]
        if len(o["commands"]) > 2:
            error(
                f"Please provide at most two command alternatives. Found {len(o['commands'])} for {func.__name__}")

        o["desc"] = kwargs.get("desc", "-")
        o["arg_template"] = kwargs.get("template")
        o["usage"] = kwargs.get("usage", "?")

        o["callback"] = func

        sig = signature(func)
        if len(sig.parameters) != 1:
            error(
                f"Expected 1 argument for function <{func.__name__}> (This is where a list of the shell arguments are passed)")
        _cmds.append(o)
        c.commands.append(o["commands"][0])
        return func
    return inner


def check_string_format(string, format_string):
    try:
        values = string.split()
        if len(values) != len(format_string.split()):
            return False
        types = [s[-1] for s in format_string.split()]
        formatted_values = []
        for i, value in enumerate(values):
            if types[i] == "d":
                formatted_values.append(int(value))
            elif types[i] == "f":
                formatted_values.append(float(value))
            elif types[i] == "s":
                formatted_values.append(str(value))
            else:
                raise ValueError("Invalid format string")
        return True
    except ValueError as e:
        return False


@shell_cmd(commands=["help", "?"], desc="Prints help information")
async def help_cmd(argv):
    help()


def help():
    print(bcolors.OKBLUE+"Available commands are")
    print(bcolors.ENDC, end='')
    for cmd in _cmds:
        helpstr = ""
        for c in cmd.get("commands"):
            helpstr += f"{c} "
        helpstr = helpstr[:-1]
        print(f"{helpstr} \t: {cmd.get('desc')}")


running = True
shell_task_ref_ = None


async def shell_task():
    global shell_task_ref_
    global running
    shell_task_ref_ = asyncio.current_task()
    running = True
    await c.read_input_stream()
    print("EXIT")
    running = False


@shell_cmd(commands=["exit", "q"], desc="Prints help information")
async def exit_cmd(argv):
    global running
    running = False
    shell_task_ref_.cancel()



from setuptools import setup, find_packages

VERSION = '1.0.1' 
DESCRIPTION = 'An asynchronous Interactive shell'
LONG_DESCRIPTION = 'An asynchronous interactive shell to interact with your program with autocompleting, history and argument validation'

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory /"shella"/ "README.md").read_text()
print(long_description)


setup(

        name="shella", 
        version=VERSION,
        author="Malte Gruber",
        author_email="<contact@maltegruber.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=["aiofiles",],
        
        keywords=['python', 'first package'],
        #https://pypi.org/classifiers/
        classifiers= [
            
            "Development Status :: 3 - Alpha",
            "Intended Audience :: End Users/Desktop",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: Linux ",
        ]
)